/*------------------------------------------------------------------------------
 * @file     EventRecorder.c
 * @brief    Event Recorder for Debug Messages
 * @version  1.0.0
 * @date     22. January 2016
 *----------------------------------------------------------------------------*/
 
/* Copyright (c) 2016 ARM LIMITED
 
   All rights reserved.
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:
   - Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   - Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
   - Neither the name of ARM nor the names of its contributors may be used
     to endorse or promote products derived from this software without
     specific prior written permission.
   *
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
   INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
   POSSIBILITY OF SUCH DAMAGE.
   ---------------------------------------------------------------------------*/

#include "RTE_Components.h"
#include CMSIS_device_header

#include <string.h>
#include "EventRecorder.h"

#if !defined(__USED)
  #define __USED __attribute__((used))
#endif
#if !defined(__WEAK)
  #define __WEAK __attribute__((weak))
#endif

/* User Configuration Start */

/* Number of Records in Event Buffer */
#ifndef EVENT_RECORD_COUNT
#define EVENT_RECORD_COUNT      16U    // Must be 2^n! (min 8)
#endif

/* User Configuration End */

/* Event Record */
typedef struct {
  union {                       // Head
    uint32_t h32;               // Head (32-bit)
    struct {
      uint8_t  seq0;            // Sequence Number 0
      uint8_t  ts_a;            // Timestamp (upper 8-bits)
      uint16_t id;              // Event ID (16-bit)
    } h;
  } head;
  uint32_t val1;                // Value 1 (32-bit)
  uint32_t val2;                // Value 2 (32-bit)
  union {                       // Tail
    uint32_t t32;               // Tail (32-bit)
    struct {
      uint8_t  ts_b[3];         // Timestamp (lower 24-bits)
      uint8_t  seq1;            // Sequence Number 1
    } t;
  } tail;
} EventRecord_t;

/* Event Buffer */
static EventRecord_t EventBuffer[EVENT_RECORD_COUNT];
static uint32_t      EventRecordIndex;

/* Event Statistics */
typedef struct {
  uint32_t records_written;
  uint32_t records_dumped;
} EventStats_t;

static EventStats_t EventStats;

/* TimeStamp */
static uint32_t TimeStamp;
static uint32_t TimeStampAddress;

/* Record Lock Bits */
static uint8_t RecordLocks[EVENT_RECORD_COUNT/8];

/* Global Event Recorder Information */
typedef struct {
  uint16_t   protocol_version;  // Protocol Version: [15..8]=major, [7..0]=minor
  // Version 1.0 specific information
  uint16_t       record_count;  // Number of Records in Event Buffer
  EventRecord_t *event_buffer;  // Pointer to Event Buffer
  EventStats_t  *event_stats;   // Pointer to Event Statistics
  uint32_t      *ts_addr;       // Pointer to TimeStamp Address
} EventRecorderInfo_t;

extern const EventRecorderInfo_t EventRecorderInfo;
__USED const EventRecorderInfo_t EventRecorderInfo =
{
  (1U << 8) | 0U,               // Protocol Version 1.0
  EVENT_RECORD_COUNT,
  &EventBuffer[0],
  &EventStats,
  &TimeStampAddress
};


static uint32_t atomic_inc32  (uint32_t *mem);
static uint32_t atomic_lock   (uint8_t  *mem, uint8_t mask);
static void     atomic_unlock (uint8_t  *mem, uint8_t mask);

#if (__CORTEX_M >= 3U)

static uint32_t atomic_inc32 (uint32_t *mem) {
  register uint32_t val;

  do {
    val = __LDREXW(mem);
  } while (__STREXW(val + 1U, mem));

  return val;
}

static uint32_t atomic_lock (uint8_t *mem, uint8_t mask) {
  register uint8_t val;

  do {
    val = __LDREXB(mem);
    if ((val & mask)) {
      __CLREX();
      return 0U;
    }
  } while (__STREXB(val | mask, mem));

  return 1U;
}

static void atomic_unlock (uint8_t *mem, uint8_t mask) {
  register uint8_t val;

  do {
    val = __LDREXB(mem);
  } while (__STREXB(val & ~mask, mem));
}

#else

static uint32_t atomic_inc32 (uint32_t *mem) {
  uint32_t irqmask;
  uint32_t ret;

  irqmask = __get_PRIMASK();
  __disable_irq();
  ret = *mem;
  *mem = ret + 1U;
  if (irqmask == 0U) {
    __enable_irq();
  }

  return ret;
}

static uint32_t atomic_lock (uint8_t *mem, uint8_t mask) {
  uint32_t irqmask;
  uint32_t ret;

  irqmask = __get_PRIMASK();
  __disable_irq();
  if (*mem & mask) {
    ret = 0U;
  } else {
    *mem |= mask;
    ret = 1U;
  }
  if (irqmask == 0U) {
    __enable_irq();
  }

  return ret;
}

static void atomic_unlock (uint8_t *mem, uint8_t mask) {
  uint32_t irqmask;

  irqmask = __get_PRIMASK();
  __disable_irq();
  *mem &= ~mask;
  if (irqmask == 0U) {
    __enable_irq();
  }

}

#endif


static uint32_t GetRecordIndex (void);
static uint32_t LockRecord   (uint32_t index);
static void     UnlockRecord (uint32_t index);
static void     IncrementRecordsWritten (void);
static void     IncrementRecordsDumped  (void);

__STATIC_INLINE uint32_t GetRecordIndex (void) {
  return (atomic_inc32(&EventRecordIndex));
}

__STATIC_INLINE uint32_t LockRecord (uint32_t n) {
  return (atomic_lock(&RecordLocks[n / 8], (uint8_t)(1U << (n & 0x07U))));
}

__STATIC_INLINE void UnlockRecord (uint32_t n) {
  atomic_unlock(&RecordLocks[n / 8], (uint8_t)(1U << (n & 0x07U)));
}

__STATIC_INLINE void IncrementRecordsWritten (void) {
  atomic_inc32(&EventStats.records_written);
}

__STATIC_INLINE void IncrementRecordsDumped (void) {
  atomic_inc32(&EventStats.records_dumped);
}


/**
  Initialize Event Recorder
 
  \return               1 - Success, 0 - Failure
*/
uint32_t Event_Initialize (void) {
  (void)TimeStamp;

  EventRecordIndex = 0U;
  memset(&EventBuffer[0], 0, sizeof(EventBuffer));
  memset(&RecordLocks[0], 0, sizeof(RecordLocks));
  EventStats.records_written = 0U;
  EventStats.records_dumped  = 0U;

  TimeStampAddress = Event_TimeStampInit();

  return 1U;
}

/**
  Initialize Event Timestamp
 
  \return               Timestamp address
*/
__WEAK uint32_t Event_TimeStampInit (void) {
#if (__CORTEX_M >= 3U)
  CoreDebug->DEMCR |= CoreDebug_DEMCR_TRCENA_Msk;
  DWT->CTRL |= DWT_CTRL_CYCCNTENA_Msk;
  return ((uint32_t)&DWT->CYCCNT);
#else
  return ((uint32_t)&TimeStamp);
#endif
}

/**
  Get Event Timestamp
 
  \return               Timestamp (32-bit)
*/
__WEAK uint32_t Event_TimeStampGet (void) {
#if (__CORTEX_M >= 3U)
  return (DWT->CYCCNT);
#else
  return (TimeStamp++);
#endif
}

/**
  Record single Item
 
  \param[in]    id    Event Identifier (16-bit)
  \param[in]    ts    Timestamp (32-bit)
  \param[in]    val1  Event Value 1 (32-bit)
  \param[in]    val2  Event Value 2 (32-bit)
  \return             1 - Success, 0 - Failure
*/
static uint32_t Event_RecordItem (uint16_t id, uint32_t ts, uint32_t val1, uint32_t val2) {
  EventRecord_t *record;
  uint32_t cnt, i, n;
  uint8_t  seq;

  for (cnt = EVENT_RECORD_COUNT; cnt; cnt--) {
    i = GetRecordIndex();
    n = i & (EVENT_RECORD_COUNT - 1U);
    if (LockRecord(n)) {
      seq = (uint8_t)(i / EVENT_RECORD_COUNT);

      record = &EventBuffer[n];
      record->tail.t.seq1 = seq;
      __DSB();

#ifdef __BIG_ENDIAN
      record->head.h32 = ((uint32_t)seq << 24U) | ((ts & 0xFF000000U) >> 8U)  | id;
#else
      record->head.h32 = ((uint32_t)id  << 16U) | ((ts & 0xFF000000U) >> 16U) | seq;
#endif
      record->val1 = val1;
      record->val2 = val2;
      __DSB();
#ifdef __BIG_ENDIAN
      record->tail.t32 = ((ts & 0x00FFFFFFU) << 8U) |  (~seq & 0xFFU);
#else
      record->tail.t32 =  (ts & 0x00FFFFFFU)        | ((~seq & 0xFFU) << 24U);
#endif

      UnlockRecord(n);

      IncrementRecordsWritten();
      return 1U;
    }
  }

  IncrementRecordsDumped();
  return 0U;
}

/**
  Record an Event (no parameters)
 
  \param[in]    id    Event Identifier  (16-bit)
  \return             1 - Success, 0 - Failure
*/
uint32_t Event_Record0 (uint16_t id) {
  return Event_Record2(id, 0U, 0U);
}

/**
  Record an Event with 1 parameter
 
  \param[in]    id    Event Identifier  (16-bit)
  \param[in]    par1  Event Parameter 1 (32-bit)
  \return             1 - Success, 0 - Failure
*/
uint32_t Event_Record1 (uint16_t id, uint32_t par1) {
  return Event_Record2(id, par1, 0U);
}

/**
  Record an Event with 2 parameters
 
  \param[in]    id    Event Identifier  (16-bit)
  \param[in]    par1  Event Parameter 1 (32-bit)
  \param[in]    par2  Event Parameter 2 (32-bit)
  \return             1 - Success, 0 - Failure
*/
uint32_t Event_Record2 (uint16_t id, uint32_t par1, uint32_t par2) {
  uint32_t ts;
  uint32_t ret;

  ts = Event_TimeStampGet();

  ret = Event_RecordItem(id, ts, par1, par2);
  if (ret == 0U) {
    return 0U;
  }

  return 1U;
}

/**
  Record an Event with 3 parameters
 
  \param[in]    id    Event Identifier  (16-bit)
  \param[in]    par1  Event Parameter 1 (32-bit)
  \param[in]    par2  Event Parameter 2 (32-bit)
  \param[in]    par3  Event Parameter 3 (32-bit)
  \return             1 - Success, 0 - Failure
*/
uint32_t Event_Record3 (uint16_t id,
                        uint32_t par1, uint32_t par2, uint32_t par3) {
  return Event_Record4(id, par1, par2, par3, 0U);
}

/**
  Record an Event with 4 parameters
 
  \param[in]    id    Event Identifier  (16-bit)
  \param[in]    par1  Event Parameter 1 (32-bit)
  \param[in]    par2  Event Parameter 2 (32-bit)
  \param[in]    par3  Event Parameter 3 (32-bit)
  \param[in]    par4  Event Parameter 4 (32-bit)
  \return             1 - Success, 0 - Failure
*/
uint32_t Event_Record4 (uint16_t id,
                        uint32_t par1, uint32_t par2, uint32_t par3, uint32_t par4) {
  uint32_t ts;
  uint32_t ret;

  ts = Event_TimeStampGet();

  ret = Event_RecordItem(id, ts, par1, par2);
  if (ret == 0U) {
    return 0U;
  }
  ret = Event_RecordItem(0U, ts, par3, par4);
  if (ret == 0U) {
    return 0U;
  }

  return 1U;
}

/**
  Record an Event with 5 parameters
 
  \param[in]    id    Event Identifier  (16-bit)
  \param[in]    par1  Event Parameter 1 (32-bit)
  \param[in]    par2  Event Parameter 2 (32-bit)
  \param[in]    par3  Event Parameter 3 (32-bit)
  \param[in]    par4  Event Parameter 4 (32-bit)
  \param[in]    par5  Event Parameter 5 (32-bit)
  \return             1 - Success, 0 - Failure
*/
uint32_t Event_Record5 (uint16_t id,
                        uint32_t par1, uint32_t par2, uint32_t par3, uint32_t par4,
                        uint32_t par5) {
  return Event_Record6(id, par1, par2, par3, par4, par5, 0U);
}

/**
  Record an Event with 6 parameters
 
  \param[in]    id    Event Identifier  (16-bit)
  \param[in]    par1  Event Parameter 1 (32-bit)
  \param[in]    par2  Event Parameter 2 (32-bit)
  \param[in]    par3  Event Parameter 3 (32-bit)
  \param[in]    par4  Event Parameter 4 (32-bit)
  \param[in]    par5  Event Parameter 5 (32-bit)
  \param[in]    par6  Event Parameter 6 (32-bit)
  \return             1 - Success, 0 - Failure
*/
uint32_t Event_Record6 (uint16_t id,
                        uint32_t par1, uint32_t par2, uint32_t par3, uint32_t par4,
                        uint32_t par5, uint32_t par6) {
  uint32_t ts;
  uint32_t ret;

  ts = Event_TimeStampGet();

  ret = Event_RecordItem(id, ts, par1, par2);
  if (ret == 0U) {
    return 0U;
  }
  ret = Event_RecordItem(0U, ts, par3, par4);
  if (ret == 0U) {
    return 0U;
  }
  ret = Event_RecordItem(0U, ts, par5, par6);
  if (ret == 0U) {
    return 0U;
  }

  return 1U;
}

/**
  Record an Event with 7 parameters
 
  \param[in]    id    Event Identifier  (16-bit)
  \param[in]    par1  Event Parameter 1 (32-bit)
  \param[in]    par2  Event Parameter 2 (32-bit)
  \param[in]    par3  Event Parameter 3 (32-bit)
  \param[in]    par4  Event Parameter 4 (32-bit)
  \param[in]    par5  Event Parameter 5 (32-bit)
  \param[in]    par6  Event Parameter 6 (32-bit)
  \param[in]    par7  Event Parameter 7 (32-bit)
  \return             1 - Success, 0 - Failure
*/
uint32_t Event_Record7 (uint16_t id,
                        uint32_t par1, uint32_t par2, uint32_t par3, uint32_t par4,
                        uint32_t par5, uint32_t par6, uint32_t par7) {
  return Event_Record8(id, par1, par2, par3, par4, par5, par6, par7, 0U);
}

/**
  Record an Event with 8 parameters
 
  \param[in]    id    Event Identifier  (16-bit)
  \param[in]    par1  Event Parameter 1 (32-bit)
  \param[in]    par2  Event Parameter 2 (32-bit)
  \param[in]    par3  Event Parameter 3 (32-bit)
  \param[in]    par4  Event Parameter 4 (32-bit)
  \param[in]    par5  Event Parameter 5 (32-bit)
  \param[in]    par6  Event Parameter 6 (32-bit)
  \param[in]    par7  Event Parameter 7 (32-bit)
  \param[in]    par8  Event Parameter 8 (32-bit)
  \return             1 - Success, 0 - Failure
*/
uint32_t Event_Record8 (uint16_t id,
                        uint32_t par1, uint32_t par2, uint32_t par3, uint32_t par4,
                        uint32_t par5, uint32_t par6, uint32_t par7, uint32_t par8) {
  uint32_t ts;
  uint32_t ret;

  ts = Event_TimeStampGet();

  ret = Event_RecordItem(id, ts, par1, par2);
  if (ret == 0U) {
    return 0U;
  }
  ret = Event_RecordItem(0U, ts, par3, par4);
  if (ret == 0U) {
    return 0U;
  }
  ret = Event_RecordItem(0U, ts, par5, par6);
  if (ret == 0U) {
    return 0U;
  }
  ret = Event_RecordItem(0U, ts, par7, par8);
  if (ret == 0U) {
    return 0U;
  }

  return 1U;
}
